import unittest
from unittest.mock import Mock, patch
from enum import Enum
from typing import Optional, List
from datetime import datetime

from waii_sdk_py import WAII
from waii_sdk_py.chat import Chat, ChatRequest, ChatResponse, ChatResponseData, ChatModule
from waii_sdk_py.query import GeneratedQuery
from waii_sdk_py.database import CatalogDefinition
from waii_sdk_py.semantic_context import GetSemanticContextResponse
from waii_sdk_py.chart import ChartGenerationResponse, ChartType

class TestChat(unittest.TestCase):
    def setUp(self):
        WAII.initialize(url="http://localhost:9859/api/")
        result = WAII.Database.get_connections()
        self.result = result
        snowflake_connector = None
        for connector in result.connectors:
            if connector.db_type == "snowflake" and "MOVIE_DB" in connector.key:
                snowflake_connector = connector
                break

        WAII.Database.activate_connection(snowflake_connector.key)

        self.chat = Chat

    def test_chat_message_query(self):

        chat_request = ChatRequest(
            ask="Show me movies data",
            streaming=False,
            parent_uuid=None,
            chart_type=None,
            modules=[ChatModule.QUERY],
            module_limit_in_response=1
        )

        result = self.chat.chat_message(chat_request)

        self.assertIsInstance(result, ChatResponse)
        self.assertIsNotNone(result.response_data)
        self.assertIsNotNone(result.response_data.query)
        self.assertTrue(result.is_new)
        self.assertIsNotNone(result.timestamp_ms)

    def test_chat_message_chart(self):
        chat_request = ChatRequest(
            ask="Show me a bar chart of movie revenue",
            streaming=False,
            parent_uuid=None,
            chart_type=ChartType.SUPERSET,
            modules=[ChatModule.CHART, ChatModule.QUERY, ChatModule.DATA],
            module_limit_in_response=2
        )

        result = self.chat.chat_message(chat_request)

        self.assertIsInstance(result, ChatResponse)
        self.assertIsNotNone(result.response)
        self.assertIsNotNone(result.response_data)
        self.assertIsNotNone(result.response_data.chart)
        self.assertIsNotNone(result.response_data.query)
    def test_chat_message_semantic_context(self):
        chat_request = ChatRequest(
            ask="What tables are available for movie data?",
            streaming=False,
            parent_uuid=None,
            chart_type=None,
            module_limit_in_response=1
        )

        result = self.chat.chat_message(chat_request)

        self.assertIsInstance(result, ChatResponse)
        self.assertIsNotNone(result.response)
        self.assertIsNotNone(result.response_data)
        self.assertIsNotNone(result.response_data.semantic_context)

    def test_chat_conversation(self):
        # Initial question
        chat_request1 = ChatRequest(
            ask="What's the rating of movies?",
            streaming=False,
            parent_uuid=None,
            modules=[ChatModule.QUERY, ChatModule.CHART, ChatModule.DATA],
            module_limit_in_response=2
        )

        result1 = self.chat.chat_message(chat_request1)

        self.assertIsInstance(result1, ChatResponse)
        self.assertIsNotNone(result1.chat_uuid)

        # Follow-up question
        chat_request2 = ChatRequest(
            ask="Now show me this data in a bar chart",
            streaming=False,
            parent_uuid=result1.chat_uuid,
            chart_type=ChartType.SUPERSET,
            modules=[ChatModule.CHART, ChatModule.QUERY, ChatModule.DATA],
            module_limit_in_response=1
        )

        result2 = self.chat.chat_message(chat_request2)

        self.assertIsInstance(result2, ChatResponse)
        self.assertIsNotNone(result2.response_data.chart)

if __name__ == '__main__':
    unittest.main()


import unittest
import pytest
from typing import Optional, List

from waii_sdk_py.my_pydantic import WaiiBaseModel

class ABaseModel(WaiiBaseModel):
    attr_a: int
    attr_b: float
    attr_c: str

class BBaseModel(WaiiBaseModel):
    attr_a: ABaseModel
    attr_b: int

class CBaseModel(BBaseModel):
    attr_c: int
    attr_d: ABaseModel

class DBaseModel(WaiiBaseModel):
    attr_a: int
    attr_b: List[ABaseModel]
    attr_c: Optional[CBaseModel] = None

class EBaseModel(WaiiBaseModel):
    attr_a: dict[str, ABaseModel] = {}


class TestWaiiBaseModel(unittest.TestCase):
    def test_extra_field(self):    
        # allows to add unknown attributes at init
        model_a = ABaseModel(attr_a=2, attr_b=3.3, attr_c='val', attr_unknown='unknown')
        with pytest.raises(ValueError):
            # should raise exception only when check_extra_fields() is called
            model_a.check_extra_fields()
        
        model_b = ABaseModel(attr_a=2, attr_b=3.3, attr_c='val')
        # should not raise exception
        model_b.check_extra_fields()
    
    def test_extra_field_recursive(self):
        model_a = ABaseModel(attr_a=2, attr_b=3.3, attr_c='val', attr_unknown='unknown')
        model_b = BBaseModel(attr_a=model_a, attr_b=3)

        # both models should raise exceptions
        with pytest.raises(ValueError):
            model_b.check_extra_fields()
        with pytest.raises(ValueError):
            model_a.check_extra_fields()
    
    def test_extra_field_kwargs(self):
        # kwargs
        model_c = ABaseModel(**{'attr_a':2, 'attr_b':3.3, 'attr_c':'val', 'attr_d':'new_val'})
        with pytest.raises(ValueError):
            model_c.check_extra_fields()
    
    def test_extra_field_subclass(self):
        model_a = ABaseModel(attr_a=2, attr_b=3.3, attr_c='val')
        # invalid model
        model_a_inv = ABaseModel(attr_a=2, attr_b=3.3, attr_c='val', unknown='unknown')
        model_c = CBaseModel(attr_a=model_a, attr_b=3, attr_c=4, attr_d=model_a_inv)

        with pytest.raises(ValueError):
            model_c.check_extra_fields()


        model_c = CBaseModel(attr_a=model_a, attr_b=3, attr_c=4, attr_d=model_a)
        # should not raise exception since all attributes are recursively correct
        model_c.check_extra_fields()
    
    def test_extra_field_list(self):
        model_d = DBaseModel(
            attr_a=2,
            attr_b=[
                ABaseModel(attr_a=2, attr_b=3.3, attr_c='a'),
                # invalid
                ABaseModel(attr_a=2, attr_b=3.3, attr_c='val', unknown='unknown')
            ],
        )

        with pytest.raises(ValueError):
            model_d.check_extra_fields()
    
    def test_extra_field_dict(self):
        model_e = EBaseModel(
            attr_a={'a': ABaseModel(attr_a=2, attr_b=2.2, attr_c='hello', attr_d=4)}
        )

        with pytest.raises(ValueError):
            model_e.check_extra_fields()

        model_e = EBaseModel(
            attr_a={'a': ABaseModel(attr_a=2, attr_b=2.2, attr_c='hello')} # no extra attribute, should not throw error
        )
        model_e.check_extra_fields()
    
    def test_extra_field_optional(self):
        model_d = DBaseModel(
            attr_a=2,
            attr_b=[
                ABaseModel(attr_a=2, attr_b=3.3, attr_c='a'),
                ABaseModel(attr_a=2, attr_b=3.3, attr_c='val')
            ],
            attr_c=CBaseModel(
                attr_a=ABaseModel(attr_a=2, attr_b=3.3, attr_c='a'),
                attr_b=2, attr_c=3,
                # invalid
                attr_d=ABaseModel(attr_a=2, attr_b=3.3, attr_c='v', unknown='unknown')
            )
        )

        with pytest.raises(ValueError):
            model_d.check_extra_fields()


if __name__ == '__main__':
    unittest.main()


import asyncio
import functools
import inspect
from typing import TypeVar, Callable, Awaitable, Any

from typing_extensions import ParamSpec

T = TypeVar('T')
P = ParamSpec('P')
def to_async(func: Callable[P, T]) -> Callable[P, Awaitable[T]]:
    """Decorator to convert a sync method to async"""
    @functools.wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        return await asyncio.get_event_loop().run_in_executor(
            None, functools.partial(func, *args, **kwargs)
        )
    return wrapper


def wrap_methods_with_async(source_class, target_class):
    for name, method in inspect.getmembers(source_class, predicate=callable):
        if not name.startswith('_') and inspect.isroutine(method):
            async_method = to_async(getattr(source_class, name))
            setattr(target_class, name, async_method)



from contextlib import contextmanager
from typing import Optional, List

from .access_rules import AccessRuleImpl, AsyncAccessRuleImpl
from .chart import ChartImpl, Chart, AsyncChartImpl
from .chat import ChatImpl, Chat, AsyncChatImpl
from .history import HistoryImpl, History, AsyncHistoryImpl
from .query import QueryImpl, Query, AsyncQueryImpl
from .database import DatabaseImpl, Database, AsyncDatabaseImpl
from .semantic_context import SemanticContextImpl, SemanticContext, AsyncSemanticContextImpl
from .settings import SettingsImpl, AsyncSettingsImpl
from .user import UserImpl, AsyncUserImpl
from .user.user_static import User
from .waii_http_client import WaiiHttpClient
import importlib.metadata
from .my_pydantic import WaiiBaseModel

GET_MODELS_ENDPOINT = "get-models"


class GetModelsRequest(WaiiBaseModel):
    pass


class ModelType(WaiiBaseModel):
    name: str
    description: Optional[str]
    vendor: Optional[str]


class GetModelsResponse(WaiiBaseModel):
    models: Optional[List[ModelType]]


class Waii:

    def __init__(self, initialize_legacy_fields: bool = False):
        self.history = None
        self.query = None
        self.database = None
        self.semantic_context = None
        self.chat = None
        self.chart = None
        self.user = None
        self.access_rules = None
        self.settings = None
        self.initialize_legacy_fields = initialize_legacy_fields
        self.http_client = None


    def initialize(self, url: str = "https://tweakit.waii.ai/api/", api_key: str = "", verbose=False):
        http_client = WaiiHttpClient(url, api_key, verbose=verbose)
        self.http_client = http_client
        self.history = HistoryImpl(http_client)
        self.query = QueryImpl(http_client)
        self.database = DatabaseImpl(http_client)
        self.semantic_context = SemanticContextImpl(http_client)
        self.chat = ChatImpl(http_client)
        self.chart = ChartImpl(http_client)
        self.user = UserImpl(http_client)
        self.access_rules = AccessRuleImpl(http_client)
        self.settings = SettingsImpl(http_client)

        if self.initialize_legacy_fields:
            self.History = self.history
            self.Query = self.query
            self.Database = self.database
            self.SemanticContext = self.semantic_context
            self.Chart = self.chart
            self.Chat = self.chat
            self.User = self.user
            Query.http_client = http_client
            History.http_client = http_client
            Database.http_client = http_client
            SemanticContext.http_client = http_client
            User.http_client = http_client
            Chat.http_client = http_client
            Chart.http_client = http_client


        conns = self.database.get_connections().connectors
        if len(conns) > 0:
            self.database.activate_connection(conns[0].key)

    @staticmethod
    def version():
        return importlib.metadata.version('waii-sdk-py')

    def get_models(self, params: GetModelsRequest = GetModelsRequest()) -> GetModelsResponse:
        return self.http_client.common_fetch(
            GET_MODELS_ENDPOINT, params, GetModelsResponse
        )

    @contextmanager
    def impersonate_user(self, user_id: str):
        try:
            self.http_client.set_impersonate_user_id(user_id)
            yield
        finally:
            self.clear_impersonation()

    def set_impersonate_user(self, user_id: str):
        self.http_client.set_impersonate_user_id(user_id)

    def clear_impersonation(self):
        self.http_client.set_impersonate_user_id('')

WAII = Waii(True)

class AsyncWaii:

    def __init__(self):
        self.http_client = None
        self.query = None
        self.database = None
        self.semantic_context = None
        self.chat = None
        self.chart = None
        self.user = None
        self.access_rules = None
        self.settings = None
        self.history = None




    async def initialize(self, url: str = "https://tweakit.waii.ai/api/", api_key: str = "", verbose=False):
        http_client = WaiiHttpClient(url, api_key, verbose=verbose)
        self.http_client = http_client
        self.query = AsyncQueryImpl(http_client)
        self.database = AsyncDatabaseImpl(http_client)
        self.semantic_context = AsyncSemanticContextImpl(http_client)
        self.chat = AsyncChatImpl(http_client)
        self.chart = AsyncChartImpl(http_client)
        self.user = AsyncUserImpl(http_client)
        self.access_rules = AsyncAccessRuleImpl(http_client)
        self.settings = AsyncSettingsImpl(http_client)
        self.history = AsyncHistoryImpl(http_client)
        result = await self.database.get_connections()

        conns = result.connectors
        if len(conns) > 0:
            await self.database.activate_connection(conns[0].key)

    @staticmethod
    def version():
        return importlib.metadata.version('waii-sdk-py')



import time

from waii_sdk_py import WAII
from waii_sdk_py.database import ModifyDBConnectionRequest, DBConnection
from waii_sdk_py.user import ListOrganizationsRequest, ListTenantsRequest, ListUsersRequest
from waii_sdk_py.waii_sdk_py import Waii


def connect_db(db_conn_to_add: DBConnection, impersonate_user=None):
    WAII.initialize(url="http://localhost:9859/api/", verbose=True)

    if impersonate_user:
        WAII.set_impersonate_user(impersonate_user)
    else:
        WAII.clear_impersonation()

    # add the database connection
    try:
        WAII.Database.modify_connections(
            ModifyDBConnectionRequest(updated=[db_conn_to_add])
        ).connectors
    except Exception as e:
        print(e)
        print("""In order to run the test, you need to run Waii locally, and setup the following
        sudo -u $USER psql postgres
        CREATE ROLE waii WITH LOGIN CREATEDB PASSWORD 'password';
        CREATE DATABASE waii_sdk_test;
        GRANT ALL PRIVILEGES ON DATABASE waii_sdk_test TO waii;
        """)

    max_wait_time = 60

    # wait till the index is finished
    while True:
        try:
            WAII.Database.activate_connection(db_conn_to_add.key)
            catalogs = WAII.Database.get_catalogs()
            if catalogs.catalogs and len(catalogs.catalogs) > 0:
                break
        except Exception as e:
            print(e)
        max_wait_time -= 1
        if max_wait_time == 0:
            raise Exception("Cannot get the catalog")
        time.sleep(1)

    WAII.Database.activate_connection(db_conn_to_add.key)
    WAII.clear_impersonation()


def load_db_conn1():
    db_connection = DBConnection(
        key="postgresql://waii@localhost:5432/waii_sdk_test",  # will be generated by the system
        db_type="postgresql",
        host="localhost",
        port=5432,
        database="waii_sdk_test",
        username="waii",
        password="password"
    )
    return db_connection


def load_db_conn2():
    db_connection = DBConnection(
        key="postgresql://waii@localhost:5432/waii_sdk_test2",  # will be generated by the system
        db_type="postgresql",
        host="localhost",
        port=5432,
        database="waii_sdk_test2",
        username="waii",
        password="password"
    )
    return db_connection


def check_table_existence(waii: Waii, table_name, should_exist, max_wait_time=30, check_interval=1):
    """
    Check if a table exists or not in the database.

    :param table_name: Name of the table to check
    :param should_exist: Boolean, True if table should exist, False if it should not exist
    :param max_wait_time: Maximum time to wait in seconds (default 30)
    :param check_interval: Time to wait between checks in seconds (default 1)
    :return: Boolean, True if the condition is met, False if timeout occurs
    """
    end_time = time.time() + max_wait_time

    while time.time() < end_time:
        # Refresh the database connection
        waii.database.refresh_db_connection()

        # Check for table existence
        result = waii.database.get_catalogs()
        table_exists = any(
            table.name.table_name.lower() == table_name.lower()
            for catalog in result.catalogs
            for schema in catalog.schemas
            for table in schema.tables
        )

        # If the condition is met, return True
        if table_exists == should_exist:
            return True

        # Wait before next check
        time.sleep(check_interval)

    # If we've reached this point, we've timed out
    return False


def org_exists(waii: Waii, org_id: str) -> bool:
    orgs = waii.user.list_orgs(ListOrganizationsRequest()).organizations
    for org in orgs:
        if org.id == org_id:
            return True
    return False


def tenant_exists(waii: Waii, org_id: str, tenant_id: str) -> bool:
    tenants = waii.user.list_tenants(ListTenantsRequest(lookup_org_id=org_id)).tenants
    for tenant in tenants:
        if tenant.id == tenant_id:
            return True
    return False


def user_exists(waii: Waii, org_id: str, user_id: str) -> bool:
    users = waii.user.list_users(ListUsersRequest(lookup_org_id=org_id)).users
    for user in users:
        if user.id == user_id:
            return True
    return False



from enum import Enum

from waii_sdk_py.waii_http_client import WaiiHttpClient
from ..common import CommonRequest
from ..my_pydantic import WaiiBaseModel
from typing import Optional, List, Dict, Any

from ..user import CommonResponse
from waii_sdk_py.utils import wrap_methods_with_async
UPDATE_PARAMETER_ENDPOINT = "update-parameter"
LIST_PARAMETER_ENDPOINT = "list-parameters"
DELETE_PARAMETER_ENDPOINT = "delete-parameter"


class DeleteParameterRequest(CommonRequest):
    parameter: str
    target_org_id: Optional[str] = None
    target_tenant_id: Optional[str] = None
    target_user_id: Optional[str] = None
    target_connection_key: Optional[str] = None


class UpdateParameterRequest(CommonRequest):
    parameter: str
    value: Any
    target_org_id: Optional[str] = None
    target_tenant_id: Optional[str] = None
    target_user_id: Optional[str] = None
    target_connection_key: Optional[str] = None


class ParameterInfo(WaiiBaseModel):
    value: Optional[Any]
    possible_values: Optional[List[Any]]


class ListParametersResponse(CommonResponse):
    parameters: Dict[str, ParameterInfo]


class Parameters(str, Enum):
    LIKED_QUERIES_ENABLED = "PUBLIC.LIKED_QUERIES.ENABLED"
    LIKED_QUERIES_LEARNING_MODE = "PUBLIC.LIKED_QUERIES.LEARNING_MODE"
    REFLECTION_ENABLED = "PUBLIC.REFLECTION.ENABLED"


class SettingsImpl:
    def __init__(self, http_client: WaiiHttpClient):
        self.http_client = http_client

    def update_parameter(
            self, params: UpdateParameterRequest
    ) -> CommonResponse:
        return self.http_client.common_fetch(
            UPDATE_PARAMETER_ENDPOINT, params, CommonResponse
        )

    def list_parameters(
            self
    ) -> ListParametersResponse:
        return self.http_client.common_fetch(
            LIST_PARAMETER_ENDPOINT, CommonRequest(), ListParametersResponse
        )

    def delete_parameter(
            self, params: DeleteParameterRequest
    ) -> CommonResponse:
        return self.http_client.common_fetch(
            DELETE_PARAMETER_ENDPOINT, params, CommonResponse
        )


class AsyncSettingsImpl:
    def __init__(self, http_client: WaiiHttpClient):
        self._settings_impl = SettingsImpl(http_client)
        wrap_methods_with_async(self._settings_impl, self)



from enum import Enum
from typing import Optional, List

from ..my_pydantic import WaiiBaseModel

from ..common import LLMBasedRequest, GetObjectRequest, AsyncObjectResponse
from ..query import GetQueryResultResponse, GeneratedQuery
from ..database import CatalogDefinition
from ..semantic_context import GetSemanticContextResponse
from ..chart import ChartGenerationResponse, ChartType
from waii_sdk_py.utils import wrap_methods_with_async
from ..waii_http_client import WaiiHttpClient

CHAT_MESSAGE_ENDPOINT = "chat-message"
SUBMIT_CHAT_MESSAGE_ENDPOINT = "submit-chat-message"
GET_CHAT_RESPONSE_ENDPOINT = "get-chat-response"


class ChatModule(str, Enum):
    DATA = "data"
    TABLES = "tables"
    QUERY = "query"
    CHART = "chart"
    CONTEXT = "context"


class ChatRequest(LLMBasedRequest):
    ask: str

    # should we streaming the output?
    # no need to support for the first implementation
    streaming: bool = False

    # link to previous conversation, pick up where conversation left off
    parent_uuid: Optional[str]

    # optional chart type, default to plotly
    chart_type: Optional[ChartType]

    modules: Optional[List[ChatModule]]

    # optional by default there is no limit
    module_limit_in_response: Optional[int]


class ChatResponseData(WaiiBaseModel):
    data: Optional[GetQueryResultResponse]
    query: Optional[GeneratedQuery]
    chart: Optional[ChartGenerationResponse]
    semantic_context: Optional[GetSemanticContextResponse]
    tables: Optional[CatalogDefinition]

class ChatResponseStep(str, Enum):
    routing_request = "Routing Request"
    generating_query = "Generating Query"
    retrieving_context = "Retrieving Context"
    retrieving_tables = "Retrieving Tables"
    running_query = "Running Query"
    generating_chart = "Generating Chart"
    preparing_result = "Preparing Result"
    completed = "Completed"


class ChatResponse(WaiiBaseModel):
    # template response
    response: Optional[str] = None
    current_step: Optional[ChatResponseStep] = None
    response_data: Optional[ChatResponseData]
    response_selected_fields: Optional[List[ChatModule]]
    is_new: Optional[bool] = False
    timestamp_ms: Optional[int]
    chat_uuid: str


class ChatImpl:

    def __init__(self, http_client: WaiiHttpClient):
        self.http_client = http_client

    def chat_message(self, params: ChatRequest) -> ChatResponse:
        return self.http_client.common_fetch(CHAT_MESSAGE_ENDPOINT, params, ChatResponse)

    def submit_chat_message(
            self, params: ChatRequest
    ) -> AsyncObjectResponse:
        return self.http_client.common_fetch(
            SUBMIT_CHAT_MESSAGE_ENDPOINT, params, AsyncObjectResponse
        )

    def get_chat_response(
            self, params: GetObjectRequest
    ) -> ChatResponse:
        return self.http_client.common_fetch(
            GET_CHAT_RESPONSE_ENDPOINT, params, ChatResponse
        )


class AsyncChatImpl:
    def __init__(self, http_client: WaiiHttpClient):
        self._chat_impl = ChatImpl(http_client)
        wrap_methods_with_async(self._chat_impl, self)


Chat = ChatImpl(WaiiHttpClient.get_instance())


import unittest
import pytest
from waii_sdk_py import WAII
from waii_sdk_py.query import *


class TestQuery(unittest.TestCase):
    def setUp(self):
        WAII.initialize(url="http://localhost:9859/api/")
        result = WAII.Database.get_connections()
        self.result = result
        pg_connector = None
        for connector in result.connectors:
            if connector.db_type == "postgresql" and connector.database == 'waii_sdk_test':
                pg_connector = connector
                break

        WAII.Database.activate_connection(pg_connector.key)

    def test_generate(self):
        params = QueryGenerationRequest(ask="How many movies are there?", use_cache=False)
        result = WAII.Query.generate(params)
        self.assertIsInstance(result, GeneratedQuery)
        assert result.uuid is not None
        assert len(result.detailed_steps) > 0
        assert len(result.query) > 0
        assert 'public.movies' in result.query.lower()
        assert len(result.tables) > 0
        assert hasattr(result.confidence_score, "confidence_value")

        params = LikeQueryRequest(query_uuid=result.uuid, liked=True)
        result = WAII.Query.like(params)
        self.assertIsInstance(result, LikeQueryResponse)

    def test_tweak(self):
        params = QueryGenerationRequest(ask = "only return 10 results", tweak_history=[Tweak(sql="select * from information_schema.tables", ask="give me all tables")])
        result = WAII.Query.generate(params)
        print(result)
        assert "limit 10" in result.query.lower()
        assert len(result.what_changed) > 0

    def test_run(self):
        params = RunQueryRequest(query="SELECT 42")
        result = WAII.Query.run(params)
        self.assertIsInstance(result, GetQueryResultResponse)
        assert len(result.column_definitions) > 0
        assert '42' in str(result.rows[0])



    def test_submit(self):
        params = RunQueryRequest(query="SELECT 42")
        result = WAII.Query.submit(params)
        self.assertIsInstance(result, RunQueryResponse)
        assert result.query_id is not None

        params = GetQueryResultRequest(query_id=result.query_id)
        result = WAII.Query.get_results(params)
        self.assertIsInstance(result, GetQueryResultResponse)
        assert '42' in str(result.rows[0])

    def test_describe(self):
        params = DescribeQueryRequest(query="SELECT 42")
        result = WAII.Query.describe(params)
        self.assertIsInstance(result, DescribeQueryResponse)
        print(result)
        assert len(result.detailed_steps) > 0
        assert len(result.summary) > 0

    def test_transcode(self):
        params = TranscodeQueryRequest(source_dialect="mysql", target_dialect="postgres", source_query="SELECT * FROM movies")
        result = WAII.Query.transcode(params)
        self.assertIsInstance(result, GeneratedQuery)
        assert len(result.query) > 0
        assert 'movies' in result.query.lower()

    def test_like_query(self):
        params = LikeQueryRequest(liked=True,ask="like test ask", query="SELECT S_STATE FROM STORE",
                                  detailed_steps=["step3", "step4"])
        result = WAII.Query.like(params)
        self.assertIsInstance(result, LikeQueryResponse)
        params = QueryGenerationRequest(ask="like test ask")
        result = WAII.Query.generate(params)
        assert result.detailed_steps == ["step3", "step4"]

    def test_error_with_unknown_fields(self):
        params = QueryGenerationRequest(ask='like test ask', q_id='unknown id')
        
        with pytest.raises(ValueError):
            result = WAII.Query.generate(params)


if __name__ == '__main__':
    unittest.main()



from enum import Enum
from typing import Optional, List, Dict, Any

from ..my_pydantic import WaiiBaseModel

class CommonRequest(WaiiBaseModel):
    tags: Optional[List[str]] = None
    parameters: Optional[Dict[str, Any]] = None


class LLMBasedRequest(WaiiBaseModel):
    tags: Optional[List[str]] = None
    parameters: Optional[Dict[str, Any]] = None
    model: Optional[str] = None
    # should we use cache?
    use_cache: Optional[bool] = True


class CommonResponse(WaiiBaseModel):
    pass


class CheckOperationStatusRequest(CommonRequest):
    op_id: str


class OperationStatus(str, Enum):
    SUCCEEDED = "succeeded"
    FAILED = "failed"
    IN_PROGRESS = "in_progress"
    NOT_EXISTS = "not_exists"


class CheckOperationStatusResponse(CommonResponse):
    op_id: str
    status: OperationStatus
    info: Optional[str] = None


class AsyncObjectResponse(CommonResponse):
    uuid: str


class GetObjectRequest(CommonRequest):
    uuid: str



from typing import Optional, List, Dict, Any

from waii_sdk_py.common import CommonRequest, CommonResponse
from waii_sdk_py.waii_http_client import WaiiHttpClient
from ..my_pydantic import WaiiBaseModel
from waii_sdk_py.utils import wrap_methods_with_async

LIST_ACCESS_KEY_ENDPOINT = "list-access-keys"
DELETE_ACCESS_KEY_ENDPOINT = "delete-access-keys"
CREATE_KEY_ENDPOINT = "create-key"
GET_USER_INFO_ENDPOINT = "get-user-info"
UPDATE_CONFIG_ENDPOINT = "update-config"
CREATE_USER_ENDPOINT = "create-user"
DELETE_USER_ENDPOINT = "delete-user"
UPDATE_USER_ENDPOINT = "update-user"
LIST_USERS_ENDPOINT = "list-users"
CREATE_TENANT_ENDPOINT = "create-tenant"
UPDATE_TENANT_ENDPOINT = "update-tenant"
DELETE_TENANT_ENDPOINT = "delete-tenant"
LIST_TENANTS_ENDPOINT = "list-tenants"
CREATE_ORG_ENDPOINT = "create-org"
UPDATE_ORG_ENDPOINT = "update-org"
DELETE_ORG_ENDPOINT = "delete-org"
LIST_ORGS_ENDPOINT = "list-orgs"


class CreateAccessKeyRequest(WaiiBaseModel):
    name: str


class AccessKey(WaiiBaseModel):
    access_key: str
    user_id: str
    description: Optional[str]
    name: Optional[str]
    created_at: Optional[int]


class GetAccessKeyResponse(WaiiBaseModel):
    access_keys: Optional[List[AccessKey]]


class GetAccessKeyRequest(WaiiBaseModel):
    pass


class DelAccessKeyRequest(WaiiBaseModel):
    names: List[str]


class DelAccessKeyResponse(WaiiBaseModel):
    pass


class GetUserInfoRequest:
    pass


class GetUserInfoResponse(WaiiBaseModel):
    id: str
    name: str
    email: str
    roles: List[str]
    permissions: List[str]


class UpdateConfigRequest(WaiiBaseModel):
    updated: Optional[Dict[str, Any]]
    deleted: Optional[List[str]]


class UpdateConfigResponse(WaiiBaseModel):
    configs: Dict[str, Any]


class Organization(WaiiBaseModel):
    id: str  # unique id for the organization
    name: str  # display name for the organization
    variables: Optional[Dict[str, Any]]  # variables for the organization


class Tenant(WaiiBaseModel):
    id: str  # unique id for the tenant
    name: str  # display name for the tenant
    org_id: Optional[str]  # org id for the tenant
    variables: Optional[Dict[str, Any]]  # variables for the tenant


class User(WaiiBaseModel):
    id: str  # unique id for the user
    name: Optional[str]  # display name for the user
    tenant_id: Optional[str]  # tenant id for the user
    org_id: Optional[str]  # org id for the user
    variables: Optional[Dict[str, Any]]  # variables for the user
    roles: Optional[List[str]] = []  # roles for the user

# Roles of the user
class WaiiRoles:
    WAII_TRIAL_USER = 'waii-trial-user'
    WAII_USER = 'waii-user'
    WAII_API_USER = 'waii-api-user'
    WAII_ADMIN_USER = 'waii-admin-user'
    WAII_ORG_ADMIN_USER = 'waii-org-admin-user'
    WAII_SUPER_ADMIN_USER = 'waii-super-admin-user'
    WAII_TRAIL_API_USER = 'waii-trial-api-user'

class CreateUserRequest(CommonRequest):
    user: User


class UpdateUserRequest(CommonRequest):
    user: User


class DeleteUserRequest(CommonRequest):
    id: str


class ListUsersRequest(CommonRequest):
    # lookup by org, by default it will lookup users within the same org
    lookup_org_id: Optional[str]

    # lookup by tenant, by default it will lookup all users within the org (depends on lookup_org_id)
    lookup_tenant_id: Optional[str]

    # lookup by user, by default it will return all users after apply the above filters
    lookup_user_id: Optional[str]


class ListUsersResponse(CommonResponse):
    users: List[User]


class CreateTenantRequest(WaiiBaseModel):
    tenant: Tenant


class UpdateTenantRequest(WaiiBaseModel):
    tenant: Tenant


class DeleteTenantRequest(WaiiBaseModel):
    id: str


class ListTenantsRequest(CommonRequest):
    lookup_org_id: Optional[str]


class ListTenantsResponse(CommonResponse):
    tenants: List[Tenant]


class CreateOrganizationRequest(WaiiBaseModel):
    organization: Organization


class UpdateOrganizationRequest(WaiiBaseModel):
    organization: Organization


class DeleteOrganizationRequest(WaiiBaseModel):
    id: str


class ListOrganizationsRequest(CommonRequest):
    pass


class ListOrganizationsResponse(CommonResponse):
    organizations: List[Organization]


class UserImpl:

    def __init__(self, http_client: WaiiHttpClient):
        self.http_client = http_client

    def create_access_key(self, params: CreateAccessKeyRequest):
        return self.http_client.common_fetch(
            CREATE_KEY_ENDPOINT, params, GetAccessKeyResponse, need_scope=False
        )

    def list_access_keys(self, params: GetAccessKeyRequest):
        return self.http_client.common_fetch(LIST_ACCESS_KEY_ENDPOINT, params, GetAccessKeyResponse, need_scope=False)

    def delete_access_key(self, params: DelAccessKeyRequest):
        return self.http_client.common_fetch(DELETE_ACCESS_KEY_ENDPOINT, params, DelAccessKeyResponse, need_scope=False)

    def get_user_info(self, params: GetUserInfoRequest):
        return self.http_client.common_fetch(GET_USER_INFO_ENDPOINT, params, GetUserInfoResponse, need_scope=False)

    def update_config(self, params: UpdateConfigRequest):
        return self.http_client.common_fetch(UPDATE_CONFIG_ENDPOINT, params, UpdateConfigResponse, need_scope=False)

    def create_user(self, params: CreateUserRequest):
        return self.http_client.common_fetch(CREATE_USER_ENDPOINT, params, CommonResponse, need_scope=False)

    def delete_user(self, params: DeleteUserRequest):
        return self.http_client.common_fetch(DELETE_USER_ENDPOINT, params, CommonResponse, need_scope=False)

    def update_user(self, params: UpdateUserRequest):
        return self.http_client.common_fetch(UPDATE_USER_ENDPOINT, params, CommonResponse, need_scope=False)

    def list_users(self, params: ListUsersRequest):
        return self.http_client.common_fetch(LIST_USERS_ENDPOINT, params, ListUsersResponse, need_scope=False)

    def create_tenant(self, params: CreateTenantRequest):
        return self.http_client.common_fetch(CREATE_TENANT_ENDPOINT, params, CommonResponse, need_scope=False)

    def update_tenant(self, params: UpdateTenantRequest):
        return self.http_client.common_fetch(UPDATE_TENANT_ENDPOINT, params, CommonResponse, need_scope=False)

    def delete_tenant(self, params: DeleteTenantRequest):
        return self.http_client.common_fetch(DELETE_TENANT_ENDPOINT, params, CommonResponse, need_scope=False)

    def list_tenants(self, params: ListTenantsRequest):
        return self.http_client.common_fetch(LIST_TENANTS_ENDPOINT, params, ListTenantsResponse, need_scope=False)

    def create_org(self, params: CreateOrganizationRequest):
        return self.http_client.common_fetch(CREATE_ORG_ENDPOINT, params, CommonResponse, need_scope=False)

    def update_org(self, params: UpdateOrganizationRequest):
        return self.http_client.common_fetch(UPDATE_ORG_ENDPOINT, params, CommonResponse, need_scope=False)

    def delete_org(self, params: DeleteOrganizationRequest):
        return self.http_client.common_fetch(DELETE_ORG_ENDPOINT, params, CommonResponse, need_scope=False)

    def list_orgs(self, params: ListOrganizationsRequest):
        return self.http_client.common_fetch(LIST_ORGS_ENDPOINT, params, ListOrganizationsResponse, need_scope=False)


class AsyncUserImpl:
    def __init__(self, http_client: WaiiHttpClient):
        self._user_impl = UserImpl(http_client)
        wrap_methods_with_async(self._user_impl, self)



from .waii_sdk_py import WAII

# Add these two imports to avoid you have to from waii_sdk_py.waii_sdk_py import AsyncWaii
# (which is ugly).
from .waii_sdk_py import AsyncWaii
from .waii_sdk_py import Waii


from enum import Enum
from typing import List, Optional

from ..chat import ChatRequest, ChatResponse
from ..my_pydantic import WaiiBaseModel
from ..query import GeneratedQuery, QueryGenerationRequest
from ..chart import ChartGenerationRequest, ChartGenerationResponse
from waii_sdk_py.utils import wrap_methods_with_async
from ..waii_http_client import WaiiHttpClient

LIST_ENDPOINT = "get-generated-query-history"
GET_ENDPOINT = "get-history"


class GeneratedHistoryEntryBase(WaiiBaseModel):
    history_type: str
    # milliseconds since epoch
    timestamp_ms: Optional[int]


class GeneratedChartHistoryEntry(GeneratedHistoryEntryBase):
    request: Optional[ChartGenerationRequest]
    response: Optional[ChartGenerationResponse]


class GeneratedChatHistoryEntry(GeneratedHistoryEntryBase):
    request: Optional[ChatRequest]
    response: Optional[ChatResponse]


class GeneratedHistoryEntryType(str, Enum):
    query = "query"
    chart = "chart"
    chat = "chat"


class GeneratedQueryHistoryEntry(GeneratedHistoryEntryBase):
    query: Optional[GeneratedQuery] = None
    request: Optional[QueryGenerationRequest] = None


class GetGeneratedQueryHistoryRequest(WaiiBaseModel):
    limit: Optional[int] = None
    offset: Optional[int] = None


class GetGeneratedQueryHistoryResponse(WaiiBaseModel):
    history: Optional[List[GeneratedQueryHistoryEntry]] = None


class SortOrder(str, Enum):
    asc = "asc"
    desc = "desc"


class GetHistoryResponse:
    def __init__(self, objs):
        self.history = []
        if 'history' not in objs:
            raise Exception(f"history is required, but not found in the response, {objs}")

        objs = objs['history']
        for h in objs:
            if 'history_type' not in h:
                raise Exception(f"history_type is required, but not found in the response, {h}")

            history_type = h['history_type']

            if history_type == GeneratedHistoryEntryType.query:
                self.history.append(GeneratedQueryHistoryEntry(**h))
            elif history_type == GeneratedHistoryEntryType.chart:
                self.history.append(GeneratedChartHistoryEntry(**h))
            elif history_type == GeneratedHistoryEntryType.chat:
                self.history.append(GeneratedChatHistoryEntry(**h))

class GetHistoryRequest(WaiiBaseModel):
    # by default include query for backward compatibility
    included_types: Optional[List[GeneratedHistoryEntryType]] = [GeneratedHistoryEntryType.query,
                                                                 GeneratedHistoryEntryType.chart,
                                                                 GeneratedHistoryEntryType.chat]

    # for pagination
    limit: Optional[int] = 1000
    offset: Optional[int] = 0

    # latest first (default)
    timestamp_sort_order: Optional[SortOrder] = SortOrder.desc

    # filter by uuid of the entry
    uuid_filter: Optional[str] = None

    # filter by the liked query flag, by default it will include both liked and unliked queries.
    # when this is set to not None, you must only include query as included_types.
    liked_query_filter: Optional[bool] = None


class HistoryImpl:
    def __init__(self, http_client: WaiiHttpClient):
        self.http_client = http_client

    # this is deprecated, use get() instead
    def list(
            self,
            params: Optional[GetGeneratedQueryHistoryRequest] = None,
    ) -> GetGeneratedQueryHistoryResponse:
        if params == None:
            params = GetGeneratedQueryHistoryRequest()
        print("This method is deprecated, use get() instead")
        return self.http_client.common_fetch(
            LIST_ENDPOINT, params, GetGeneratedQueryHistoryResponse
        )

    def get(
            self,
            params: Optional[GetHistoryRequest] = None,
    ) -> GetHistoryResponse:
        if params == None:
            params = GetHistoryRequest()
        objs = self.http_client.common_fetch(
            GET_ENDPOINT, params, ret_json=True
        )
        return GetHistoryResponse(objs)


class AsyncHistoryImpl:
    def __init__(self, http_client: WaiiHttpClient):
        self._history_impl = HistoryImpl(http_client)
        wrap_methods_with_async(self._history_impl, self)


History = HistoryImpl(WaiiHttpClient.get_instance())



import asyncio
import time
import unittest
from unittest import IsolatedAsyncioTestCase

from tests.common_test_utils import load_db_conn1
from waii_sdk_py.database import TableDefinition, TableName, ColumnDefinition, UpdateTableDefinitionRequest, \
    ModifyDBConnectionRequest, TableReference
from waii_sdk_py.query import QueryGenerationRequest, GeneratedQuery
from waii_sdk_py.waii_sdk_py import AsyncWaii


async def generate_task(method, params, task_name, task_indicators, start_times):
    start_times.append((task_name, time.time()))
    task_indicators.append(f"{task_name}_started")

    result = await method(params) if params else await method()

    task_indicators.append(f"{task_name}_finished")
    start_times.append((task_name, time.time()))

    return result


class TestAsyncQuery(IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        async_waii_client = AsyncWaii()
        await async_waii_client.initialize(url="http://localhost:9859/api/")
        self.db_conn = load_db_conn1()
        self.async_waii_client = async_waii_client
        result = await self.async_waii_client.database.get_connections()
        self.result = result
        pg_connector = None
        for connector in result.connectors:
            if connector.db_type == "postgresql":
                pg_connector = connector
                break

        await self.async_waii_client.database.activate_connection(pg_connector.key)

    async def test_generate_query(self):
        params = QueryGenerationRequest(ask="List all tables in database")

        start = time.time()
        result = await self.async_waii_client.query.generate(params)
        isolated_execution_time = time.time() - start

        task_indicators = []
        start_times = []
        params = QueryGenerationRequest(ask="List all tables in database?.")
        task1 = asyncio.create_task(generate_task(
            self.async_waii_client.query.generate, params, "task1", task_indicators, start_times
        ))
        params = QueryGenerationRequest(ask="List all tables in database?...")
        task2 = asyncio.create_task(generate_task(
            self.async_waii_client.query.generate, params, "task2", task_indicators, start_times
        ))

        concurrent_start = time.time()
        await asyncio.gather(task1, task2)
        concurrent_execution_time = time.time() - concurrent_start
        print(f"Isolated execution time: {isolated_execution_time}")
        print(f"Concurrent execution time: {concurrent_execution_time}")
        self.assertLess(concurrent_execution_time, isolated_execution_time * 1.5,
                        "Tasks did not run concurrently, likely due to blocking HTTP client")

        # Validate order of task start and end times
        task1_start, task2_start = start_times[0][1], start_times[1][1]


        # Expect both tasks to overlap significantly in start times for async behavior
        self.assertLess(abs(task1_start - task2_start), 0.1, "Tasks should start concurrently")

        self.assertEqual("task1_started", task_indicators[0])
        self.assertEqual("task2_started", task_indicators[1])
        self.assertIn("task1_finished", task_indicators)
        self.assertIn("task2_finished", task_indicators)
        self.assertIsInstance(result, GeneratedQuery)
        assert result.uuid is not None
        assert len(result.detailed_steps) > 0
        assert len(result.query) > 0
        assert 'information_schema' in result.query.lower()
        assert len(result.tables) > 0
        assert hasattr(result.confidence_score, "confidence_value")

    async def test_modify_connections(self):
        # first try to delete te connection if it exists
        result = await self.async_waii_client.database.get_connections()
        result = result.connectors
        for conn in result:
            if (
                    conn.db_type == self.db_conn.db_type
                    and conn.account_name == self.db_conn.account_name
                    and self.db_conn.database == conn.database
                    and self.db_conn.username == conn.username
            ):
                result = await self.async_waii_client.database.modify_connections(
                    ModifyDBConnectionRequest(removed=[conn.key])
                )
                result = result.connectors
                break

        # then add the new connection
        new_result = await self.async_waii_client.database.modify_connections(
            ModifyDBConnectionRequest(updated=[self.db_conn])
        )
        new_result = new_result.connectors

        assert len(result) == len(new_result) - 1

        # then activate the new connection
        for conn in new_result:
            if (
                    conn.db_type == self.db_conn.db_type
                    and conn.account_name == self.db_conn.account_name
                    and self.db_conn.database == conn.database
                    and self.db_conn.username == conn.username
            ):
                await self.async_waii_client.database.activate_connection(conn.key)
                break
        else:
            raise Exception("Cannot find the new connection")

        # list databases
        result = await self.async_waii_client.database.get_catalogs()
        assert len(result.catalogs) > 0
        for catalog in result.catalogs:
            for schema in catalog.schemas:
                for table in schema.tables:
                    if len(table.refs) > 0:
                        assert type(table.refs[0]) == TableReference









import inspect
import json
import warnings

from waii_sdk_py.waii_http_client import WaiiHttpClient
from ..common import LLMBasedRequest, CommonRequest, CheckOperationStatusResponse, CheckOperationStatusRequest
from ..my_pydantic import WaiiBaseModel, PrivateAttr
import re
from typing import Optional, List, Dict, Any, Union
from urllib.parse import urlparse, parse_qs
from enum import Enum

from ..user import CommonResponse
from ..utils.utils import to_async, wrap_methods_with_async

MODIFY_DB_ENDPOINT = "update-db-connect-info"
GET_CATALOG_ENDPOINT = "get-table-definitions"
GET_MODELS_ENDPOINT = "get-models"
UPDATE_TABLE_DESCRIPTION_ENDPOINT = "update-table-description"
UPDATE_SCHEMA_DESCRIPTION_ENDPOINT = "update-schema-description"
UPDATE_COLUMN_DESCRIPTION_ENDPOINT = "update-column-description"
UPDATE_CONSTRAINT_ENDPOINT = "update-constraint"
UPDATE_TABLE_DEFINITION_ENDPOINT = "update-table-definitions"
UPDATE_SIMILARITY_SEARCH_INDEX_ENDPOINT = "update-similarity-search-index"
GET_SIMILARITY_SEARCH_INDEX_ENDPOINT = "get-similarity-search-index"
DELETE_SIMILARITY_SEARCH_INDEX_ENDPOINT = "delete-similarity-search-index"
GET_SIMILARITY_SEARCH_INDEX_TABLE_ENDPOINT = "get-similarity-search-index-table"
CHECK_SIMILARITY_SEARCH_INDEX_STATUS_ENDPOINT = "check-similarity-search-index-status"


class SchemaName(WaiiBaseModel):
    schema_name: str
    database_name: Optional[str]


class TableName(WaiiBaseModel):
    table_name: str
    schema_name: Optional[str]
    database_name: Optional[str]


class ColumnName(WaiiBaseModel):
    table_name: TableName
    column_name: str


class ColumnSampleValues(WaiiBaseModel):
    values: Optional[Dict[str, int]]


class ColumnDefinition(WaiiBaseModel):
    name: str
    type: str
    comment: Optional[str]
    description: Optional[str]
    sample_values: Optional[ColumnSampleValues]

    def __lt__(self, other):
        return self.name < other.name


class ConstraintType(str, Enum):
    primary = "primary"
    foreign = "foreign"


class ConstraintDetectorType(str, Enum):
    database = "database"
    inferred_llm = "inferred_llm"
    inferred_static = "inferred_static"
    inferred_query_history = "inferred_query_history"
    manual = "manual"


class TableReference(WaiiBaseModel):
    src_table: Optional[TableName]  # table name
    src_cols: Optional[List[str]]  # source table columns
    ref_table: Optional[TableName]  # ref table name
    ref_cols: Optional[List[str]]  # ref table columns
    source: Optional[ConstraintDetectorType]


class TableNameToDescription(WaiiBaseModel):
    name: str
    description: str


class SchemaDescription(WaiiBaseModel):
    summary: Optional[str]
    common_questions: Optional[List[str]]
    common_tables: Optional[List[TableNameToDescription]]


class RelationshipType(str, Enum):
    one_to_one = "one_to_one"
    one_to_many = "one_to_many"
    many_to_many = "many_to_many"
    belongs_to = "belongs_to"
    has_one = "has_one"
    has_many = "has_many"
    many_to_one = "many_to_one"


class Constraint(WaiiBaseModel):
    source: Optional[ConstraintDetectorType]
    table: Optional[TableName]
    cols: Optional[List[str]]
    constraint_type: Optional[ConstraintType]
    relationship_type: Optional[RelationshipType]

    # for foreign key, it is the table and columns that the source table and cols
    src_table: Optional[TableName]  # table name
    src_cols: Optional[List[str]]  # source table columns

    # comment for the constraint
    comment: Optional[str]

    def __repr__(self):
        if self.constraint_type == ConstraintType.primary:
            return f"PK {self.table} ({self.cols} [{self.source.name}]"
        else:
            return f"FK {self.src_table} ({self.src_cols}) -> {self.table} ({self.cols}) [{self.source.name}]"


class TableDefinition(WaiiBaseModel):
    name: TableName
    columns: Optional[List[ColumnDefinition]]
    comment: Optional[str]
    last_altered_time: Optional[int]
    _refs: Optional[List[TableReference]] = PrivateAttr(default=[])
    constraints: Optional[List[Constraint]]
    inferred_refs: Optional[List[TableReference]]
    inferred_constraints: Optional[List[Constraint]]
    description: Optional[str]

    ddl: Optional[str] = None

    def __init__(self, **data):
        super().__init__(**data)
        refs_data = data.get("refs",None)
        refs = []
        if refs_data:
            for ref in refs_data:
                if type(ref) == TableReference:
                    refs.append(ref)
                else:
                    refs.append(TableReference(**ref))

        self._refs = refs

    @property
    def refs(self):
        warnings.warn(
            "The 'refs' attribute is deprecated and will be removed in a future release. Use 'constraints' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._refs

    @refs.setter
    def refs(self, value):
        warnings.warn(
            "The 'refs' attribute is deprecated and will be removed in a future release. Use 'constraints' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self._refs = value

class SchemaDefinition(WaiiBaseModel):
    name: SchemaName
    tables: Optional[List[TableDefinition]]
    description: Optional[SchemaDescription]


class CatalogDefinition(WaiiBaseModel):
    name: str
    schemas: Optional[List[SchemaDefinition]]

class DBContentFilterScope(str, Enum):
    schema = "schema"
    table = "table"
    column = "column"

class DBContentFilterType(str, Enum):
    include = "include"
    exclude = "exclude"

class DBContentFilterActionType(str, Enum):
    # this means we will include/exclude the content
    visibility = "visibility"

    # this means we will do sample value or not
    sample_values = "sample_values"


class FilterType(str, Enum):
    INCLUSION = "inclusion"
    EXCLUSION = "exclusion"


class SearchContext(WaiiBaseModel):
    type: Optional[FilterType] = FilterType.INCLUSION
    db_name: Optional[str] = '*'
    schema_name: Optional[str] = '*'
    table_name: Optional[str] = '*'
    column_name: Optional[str] = '*'
    ignore_case: Optional[bool] = True

class DBContentFilter(WaiiBaseModel):
    filter_scope: DBContentFilterScope
    filter_type: DBContentFilterType = DBContentFilterType.include
    filter_action_type: DBContentFilterActionType = DBContentFilterActionType.visibility
    ignore_case: Optional[bool] = True
    pattern: str  # regex pattern
    search_context: Optional[List[SearchContext]]


class DBAccessPolicy(WaiiBaseModel):
    read_only: Optional[bool] = False
    allow_access_beyond_db_content_filter: Optional[bool] = True
    allow_access_beyond_search_context: Optional[bool] = True

class DBConnection(WaiiBaseModel):
    key: Optional[str] = None
    db_type: str
    description: Optional[str] = None
    account_name: Optional[str] = None
    username: Optional[str] = None
    password: Optional[str] = None
    database: Optional[str] = None
    warehouse: Optional[str] = None
    role: Optional[str] = None
    path: Optional[str] = None
    host: Optional[str] = None
    port: Optional[int] = None
    parameters: Optional[Dict[str, Any]] = None
    sample_col_values: Optional[bool] = True
    push:Optional[bool] = False
    db_content_filters: Optional[List[DBContentFilter]] = None
    embedding_model: Optional[str] = None
    always_include_tables: Optional[List[TableName]] = None
    alias: Optional[str] = None
    db_access_policy: Optional[DBAccessPolicy] = DBAccessPolicy()
    host_alias : Optional[str] = None
    user_alias: Optional[str] = None
    db_alias: Optional[str] = None
    client_email: Optional[str] = None
    content_filters: Optional[List[SearchContext]] = None
    sample_filters: Optional[List[SearchContext]] = None

class ModifyDBConnectionRequest(WaiiBaseModel):
    updated: Optional[List[DBConnection]] = None
    removed: Optional[List[str]] = None
    validate_before_save: Optional[bool] = None
    user_id: Optional[str] = None
    default_db_connection_key: Optional[str] = None
    owner_user_id: Optional[str] = None


class SchemaIndexingStatus(WaiiBaseModel):
    n_pending_indexing_tables: int
    n_total_tables: int
    status: str


class DBConnectionIndexingStatus(WaiiBaseModel):
    status: Optional[str]
    schema_status: Optional[Dict[str, SchemaIndexingStatus]]


class ModifyDBConnectionResponse(CommonResponse):
    connectors: Optional[List[DBConnection]]
    diagnostics: Optional[str]
    default_db_connection_key: Optional[str]
    connector_status: Optional[Dict[str, DBConnectionIndexingStatus]]


class GetCatalogRequest(LLMBasedRequest):
    search_context: Optional[List[SearchContext]] = None

    # ask is to perform semantic search, if it is provided, then we will use it to search the table definition
    # we will try to find the same tables which can answer the ask
    ask: Optional[str] = None

    internal: bool = False


class GetDBConnectionRequest(CommonRequest):
    pass


class GetDBConnectionResponse(CommonResponse):
    connectors: Optional[List[DBConnection]]
    diagnostics: Optional[str] = None
    default_db_connection_key: Optional[str] = None
    connector_status: Optional[Dict[str, DBConnectionIndexingStatus]] = None

class GetCatalogResponse(CommonResponse):
    catalogs: Optional[List[CatalogDefinition]] = None
    debug_info: Optional[Dict[str, Any]]


class UpdateTableDescriptionRequest(WaiiBaseModel):
    table_name: TableName
    description: str


class UpdateTableDefinitionRequest(WaiiBaseModel):
    updated_tables: Optional[List[TableDefinition]]
    removed_tables: Optional[List[TableName]]


class ColumnDescription(WaiiBaseModel):
    column_name: str
    description: Optional[str]


class TableToColumnDescription(WaiiBaseModel):
    table_name: TableName
    column_descriptions: Optional[List[ColumnDescription]]


class UpdateColumnDescriptionRequest(WaiiBaseModel):
    col_descriptions: Optional[List[TableToColumnDescription]]


class UpdatedTableToCol(WaiiBaseModel):
    table_name: TableName
    column_names: Optional[List[str]]


class UpdateColumnDescriptionResponse(CommonResponse):
    updated_table_to_cols: Optional[List[UpdatedTableToCol]]

class UpdateTableDefinitionResponse(CommonResponse):
    updated_tables: Optional[List[TableName]]


class UpdateSchemaDescriptionRequest(WaiiBaseModel):
    schema_name: SchemaName
    description: str


class UpdateTableDescriptionResponse(CommonResponse):
    pass


class UpdateSchemaDescriptionResponse(CommonResponse):
    pass


class TableConstraints(WaiiBaseModel):
    table_name: TableName
    constraints: Optional[List[Constraint]]
    constraint_type: ConstraintType


class UpdateConstraintRequest(WaiiBaseModel):
    # updated constraints, it will replace the existing constraints
    updated_constraints: Optional[List[TableConstraints]]


class UpdateConstraintResponse(CommonResponse):
    updated_tables: Optional[List[TableName]]


class RefreshDBConnectionRequest(WaiiBaseModel):
    db_conn_key: str


class ColumnValue(WaiiBaseModel):
    value: str
    additional_info: Optional[List[str]]


class UpdateSimilaritySearchIndexRequest(CommonRequest):
    values: Optional[List[ColumnValue]]
    column: ColumnName
    enable_llm_rerank: Optional[bool] = True
    similarity_score_threshold: Optional[float] = None
    max_matched_values: Optional[int] = None
    min_matched_values: Optional[int] = None


class UpdateSimilaritySearchIndexResponse(CommonResponse):
    op_id: str


class GetSimilaritySearchIndexOnTableRequest(CommonRequest):
    table: TableName


class GetSimilaritySearchIndexOnTableResponse(CommonResponse):
    table: TableName
    columns: List[ColumnName]


class DeleteSimilaritySearchIndexRequest(CommonRequest):
    column: ColumnName

class DeleteSimilaritySearchIndexResponse(CommonResponse):
    op_id: Optional[str]


class GetSimilaritySearchIndexRequest(CommonRequest):
    column: ColumnName


class GetSimilaritySearchIndexResponse(CommonRequest):
    column: ColumnName
    values: Optional[List[ColumnValue]]


class GetModelsRequest(CommonRequest):
    pass


class Model(WaiiBaseModel):
    name: str
    description: Optional[str] = None
    vendor: Optional[str] = None


class GetModelsResponse(CommonResponse):
    models: Optional[List[Model]] = None


class DatabaseImpl:

    def __init__(self, http_client: WaiiHttpClient):
        self.http_client = http_client

    def modify_connections(
        self, params: ModifyDBConnectionRequest
    ) -> ModifyDBConnectionResponse:
        if params.updated is not None:
            self._modify_connection_request(params.updated)
        return self.http_client.common_fetch(
            MODIFY_DB_ENDPOINT, params, ModifyDBConnectionResponse, need_scope=False
        )

    def _modify_connection_request(self, conns: List[DBConnection]):
        for conn in conns:
            if conn.db_type == "bigquery" and conn.password:
                password_dict = json.loads(conn.password)
                conn.database = password_dict['project_id']
                conn.client_email = password_dict['client_email']

    def get_connections(
            self, params: Optional[GetDBConnectionRequest] = None
    ) -> GetDBConnectionResponse:
        if params == None:
            params = GetDBConnectionRequest()
        return self.http_client.common_fetch(
            MODIFY_DB_ENDPOINT,
            params,
            GetDBConnectionResponse,
            need_scope=False,
        )

    def activate_connection(self, key: str):
        # first try to get all connections and check if the key is valid
        connections = self.get_connections()

        all_connection_keys = set([conn.key for conn in connections.connectors])

        if key not in all_connection_keys:
            all_conn_keys_str = ''
            for k in all_connection_keys:
                all_conn_keys_str += f' - {k}\n'

            raise Exception(f"Connection key {key} is not valid, all the available keys are {all_conn_keys_str}.\nYou can use `WAII.Database.get_connections()` to get all the available connections.")

        self.http_client.set_scope(key)

    def get_activated_connection(self):
        return self.http_client.get_scope()

    def get_default_connection(self):
        return self.http_client.get_scope()

    def get_catalogs(
        self, params: Optional[GetCatalogRequest] = None
    ) -> GetCatalogResponse:
        if params == None:
            params = GetCatalogRequest()
        return self.http_client.common_fetch(
            GET_CATALOG_ENDPOINT, params, GetCatalogResponse
        )

    def update_table_description(
        self, params: UpdateTableDescriptionRequest
    ) -> UpdateTableDescriptionResponse:
        return self.http_client.common_fetch(
            UPDATE_TABLE_DESCRIPTION_ENDPOINT, params, GetCatalogResponse
        )

    def update_table_definition(
        self, params:UpdateTableDefinitionRequest
    ) -> UpdateTableDefinitionResponse:
        return self.http_client.common_fetch(
            UPDATE_TABLE_DEFINITION_ENDPOINT, params,UpdateTableDefinitionResponse
        )

    def update_schema_description(
        self, params: UpdateSchemaDescriptionRequest
    ) -> UpdateSchemaDescriptionResponse:
        return self.http_client.common_fetch(
            UPDATE_SCHEMA_DESCRIPTION_ENDPOINT, params, GetCatalogResponse
        )

    def update_column_description(
        self, params: UpdateColumnDescriptionRequest
    ) -> UpdateColumnDescriptionResponse:
        return self.http_client.common_fetch(
            UPDATE_COLUMN_DESCRIPTION_ENDPOINT,
            params,
            UpdateColumnDescriptionResponse,
        )

    def update_constraint(
        self, params: UpdateConstraintRequest
    ) -> UpdateConstraintResponse:
        return self.http_client.common_fetch(
            UPDATE_CONSTRAINT_ENDPOINT, params, UpdateConstraintResponse
        )

    def refresh_db_connection(self):
        return self.http_client.common_fetch(
            "refresh-db-connection",
            {
                "db_conn_key": self.get_activated_connection(),
            },
            CommonResponse, need_scope=False
        )

    def update_similarity_search_index(
            self, params: UpdateSimilaritySearchIndexRequest
    ) -> CommonResponse:
        return self.http_client.common_fetch(
            UPDATE_SIMILARITY_SEARCH_INDEX_ENDPOINT, params, UpdateSimilaritySearchIndexResponse
        )

    def get_similarity_search_index(
            self, params: GetSimilaritySearchIndexRequest
    ) -> GetSimilaritySearchIndexResponse:
        return self.http_client.common_fetch(
            GET_SIMILARITY_SEARCH_INDEX_ENDPOINT, params, GetSimilaritySearchIndexResponse
        )

    def delete_similarity_search_index(
            self, params: DeleteSimilaritySearchIndexRequest
    ) -> CommonResponse:
        return self.http_client.common_fetch(
            DELETE_SIMILARITY_SEARCH_INDEX_ENDPOINT, params, DeleteSimilaritySearchIndexResponse
        )

    def get_similarity_search_index_on_table(
            self, params: GetSimilaritySearchIndexOnTableRequest
    ) -> GetSimilaritySearchIndexOnTableResponse:
        return self.http_client.common_fetch(
            GET_SIMILARITY_SEARCH_INDEX_TABLE_ENDPOINT, params, GetSimilaritySearchIndexOnTableResponse
        )

    def get_similarity_search_index_status(
            self, params: CheckOperationStatusRequest
    ) -> CheckOperationStatusResponse:
        return self.http_client.common_fetch(
            CHECK_SIMILARITY_SEARCH_INDEX_STATUS_ENDPOINT, params, CheckOperationStatusResponse
        )

    def get_models(
            self, params: GetModelsRequest = GetModelsRequest()
    ) -> GetModelsResponse:
        return self.http_client.common_fetch(
            GET_MODELS_ENDPOINT, params, GetModelsResponse, need_scope=False
        )


class AsyncDatabaseImpl:
    def __init__(self, http_client: WaiiHttpClient):
        self._database_impl = DatabaseImpl(http_client)
        wrap_methods_with_async(self._database_impl, self)



Database = DatabaseImpl(WaiiHttpClient.get_instance())



try:
    from pydantic.v1 import (
        BaseModel,
        ValidationError,
        PrivateAttr,
        Field,
        # Add other necessary imports here
    )

    class WaiiBaseModel(BaseModel, extra='allow'):
        def check_extra_fields(self):
            extra_fields = {k: v for k, v in self.__dict__.items() if k not in self.__class__.__fields__}
            if extra_fields != {}:
                raise ValueError(f'Cannot set unknown fields: {list(extra_fields.keys())}')
            for field_name, field in self.__fields__.items():
                field_value = getattr(self, field_name)
                if field_value is None:
                    continue
                if isinstance(field_value, list):
                    for item in field_value:
                        if isinstance(item, WaiiBaseModel):
                            item.check_extra_fields()
                elif isinstance(field_value, dict):
                    for k, v in field_value.items():
                        if isinstance(v, WaiiBaseModel):
                            v.check_extra_fields()
                elif isinstance(field_value, WaiiBaseModel):
                    field_value.check_extra_fields()
        
except ImportError:
    try:
        from pydantic import (
            BaseModel,
            ValidationError,
            PrivateAttr,
            Field,
            # Add other necessary imports here
        )

        class WaiiBaseModel(BaseModel, extra='allow'):
            def check_extra_fields(self):
                if self.model_extra != {}:
                    raise ValueError(f'Cannot set unknown fields: {list(self.model_extra.keys())}')
                
                for field_name, field in self.model_fields.items():
                    field_value = getattr(self, field_name)
                    if field_value is None:
                        continue
                    if isinstance(field_value, list):
                        for item in field_value:
                            if isinstance(item, WaiiBaseModel):
                                item.check_extra_fields()
                    elif isinstance(field_value, dict):
                        for k, v in field_value.items():
                            if isinstance(v, WaiiBaseModel):
                                v.check_extra_fields()
                    elif isinstance(field_value, WaiiBaseModel):
                        field_value.check_extra_fields()

                    
    except ImportError:
        raise ImportError("Cannot find pydantic module. Please install pydantic. You can use >= 1.10.x or >= 2.7.x")

__all__ = [
    "BaseModel",
    "WaiiBaseModel",
    "ValidationError",
    "PrivateAttr",
    "Field"
]



from typing import List, Optional

from ..common import LLMBasedRequest
from ..database import SearchContext
from ..my_pydantic import WaiiBaseModel
from waii_sdk_py.utils import wrap_methods_with_async
from ..waii_http_client import WaiiHttpClient

MODIFY_ENDPOINT = 'update-semantic-context'
GET_ENDPOINT = 'get-semantic-context'


class SemanticStatement(WaiiBaseModel):
    id: Optional[str]
    statement: str
    labels: Optional[List[str]]
    scope: Optional[str]

    # always include this statement in the context, if data_type is specified in statement, then always_include is True
    # when data_type matches
    always_include: Optional[bool] = True

    # Check the application of the rule in a second step after query gen when the rule is within scope
    critical: Optional[bool] = False

    # Search keys for this statement, if not specified, then use statement as search key
    # you can specify multiple search keys, for example, if you have a CVE doc, you can search by CVE number, or library
    # name, etc.
    lookup_summaries: Optional[List[str]]

    # extract prompt from the statement, if not specified, then use statement as extract prompt
    summarization_prompt: Optional[str] = None

    # filters for user, tenant and org
    user_id: Optional[str] = '*'
    tenant_id: Optional[str] = '*'
    org_id: Optional[str] = '*'


class ModifySemanticContextRequest(LLMBasedRequest):
    updated: Optional[List[SemanticStatement]] = None
    deleted: Optional[List[str]] = None


class GetSemanticContextRequestFilter(WaiiBaseModel):
    # do we want to filter "always_include" rules or not?
    # - None: both
    # - True: only return rules with always_include=True
    # - False: only return rules with always_include=False
    always_include: Optional[bool] = None

    # Filter by labels, scope, statement.
    # They are connected by "AND", and we use substring match for them
    # Match is case insensitive
    labels: Optional[List[str]] = None  # labels to filter the rules
    scope: Optional[str] = None  # scope to filter the rules
    statement: Optional[str] = None  # statement to filter the rules

class ModifySemanticContextResponse(WaiiBaseModel):
    updated: Optional[List[SemanticStatement]] = None
    deleted: Optional[List[str]] = None


class GetSemanticContextRequest(LLMBasedRequest):
    filter: GetSemanticContextRequestFilter = GetSemanticContextRequestFilter()
    offset: int = 0
    limit: int = 1000
    search_text: Optional[str] = None
    search_context: Optional[List[SearchContext]] = None


class GetSemanticContextResponse(WaiiBaseModel):
    semantic_context: Optional[List[SemanticStatement]] = None

    available_statements: Optional[int] = 0


class SemanticContextImpl:

    def __init__(self, http_client: WaiiHttpClient):
        self.http_client = http_client

    def modify_semantic_context(self,params: ModifySemanticContextRequest) -> ModifySemanticContextResponse:
        return self.http_client.common_fetch(MODIFY_ENDPOINT, params, ModifySemanticContextResponse)


    def get_semantic_context(self, params: Optional[GetSemanticContextRequest] = None) -> GetSemanticContextResponse:
        if params == None:
            params = GetSemanticContextRequest()
        return self.http_client.common_fetch(GET_ENDPOINT, params, GetSemanticContextResponse)


class AsyncSemanticContextImpl:
    def __init__(self, http_client: WaiiHttpClient):
        self._semantic_context_impl = SemanticContextImpl(http_client)
        wrap_methods_with_async(self._semantic_context_impl, self)


SemanticContext = SemanticContextImpl(WaiiHttpClient.get_instance())


import unittest
import pytest
from waii_sdk_py import WAII
from waii_sdk_py.history import *
from waii_sdk_py.query import RunQueryRequest


class TestHistory(unittest.TestCase):
    def setUp(self):
        WAII.initialize(url="http://localhost:9859/api/")
        result = WAII.Database.get_connections()
        self.result = result
        WAII.Database.activate_connection(result.connectors[0].key)

    def test_list_legacy_api(self):
        # do a list and see if it returns something
        result = WAII.History.list()
        if not result.history:
            # first we need to create some queries
            result = WAII.Query.generate(QueryGenerationRequest(ask='how many tables are there in the database', parent_uuid=None))
            WAII.Query.run(RunQueryRequest(query=result.query))

        # Call the function
        result = WAII.History.list()
        # Check the result
        # Note: The specifics of this assertion would depend on what the function should return
        self.assertGreater(len(result.history), 0)

        params = GetGeneratedQueryHistoryRequest(limit=1, offset=1)
        result = WAII.History.get(params)
        self.assertTrue(len(result.history) >= 1)

    def test_get_new_api(self):
        result = WAII.History.get(GetHistoryRequest())
        if not result.history:
            # Then generate some chats
            asks = [
                "tell me how many tables are there in the database",
                "list number of columns for each table"
            ]
            parent_uuid = None
            for a in asks:
                chat_request = ChatRequest(ask=a, parent_uuid=parent_uuid)
                chat_response = WAII.Chat.chat_message(chat_request)
                print('Ask:', a)
                print(chat_response.response)
                if chat_response.response_data.sql:
                    print(chat_response.response_data.sql.query)
                if chat_response.response_data.data:
                    print(f"There're {len(chat_response.response_data.data.rows)} rows")
                parent_uuid = chat_response.chat_uuid

        # Then use the get method
        result = WAII.History.get(GetHistoryRequest())

        # Check the result
        self.assertGreater(len(result.history), 0)

        total_result = len(result.history)

        # Test with limit
        result = WAII.History.get(GetHistoryRequest(limit=1))
        self.assertEqual(len(result.history), 1)

        # Test with offset
        result = WAII.History.get(GetHistoryRequest(offset=1))
        self.assertEqual(len(result.history), total_result - 1)

        # Test error with unknown fields
        with pytest.raises(ValueError):
            result = WAII.History.get(
                GetHistoryRequest(offset=1, lt=5, id=4) # lt and id fields don't exist
            )



if __name__ == '__main__':
    unittest.main()



import unittest
import pytest
from waii_sdk_py import WAII
from waii_sdk_py.semantic_context import GetSemanticContextRequest

class TestSemanticContext(unittest.TestCase):
    def setUp(self):
        WAII.initialize(url="http://localhost:9859/api/")
        result = WAII.Database.get_connections()
        self.result = result
        WAII.Database.activate_connection(result.connectors[2].key)

    def test_modify_semantic_context(self):
        # Define test parameters
        # params = ModifySemanticContextRequest(updated=[SemanticStatement(statement="Example statement")])
        # Call the function
        # result = self.semantic_context.modifySemanticContext(params)
        # Check the result
        # Note: The specifics of this assertion would depend on what the function should return
        # self.assertIsInstance(result, ModifySemanticContextResponse)
        # self.assertEqual(len(result.updated), 1)
        # self.assertEqual(result.updated[0].statement, "Example statement")
        pass

    def test_get_semantic_context(self):
        # Call the function
        result = WAII.SemanticContext.get_semantic_context()
        # Check the result
        # Note: The specifics of this assertion would depend on what the function should return
        self.assertGreater(len(result.semantic_context), 0)
    
    def test_unknown_fields_error(self):
        with pytest.raises(ValueError):
            result = WAII.SemanticContext.get_semantic_context(
                GetSemanticContextRequest(search_txt='val') # search_txt field doesn't exist
            )

if __name__ == '__main__':
    unittest.main()
